<?php

/* session_2.php: Additional functions for settings, contacts, etc.
*
    Copyright (C) 2002-2005  Hastymail Development group

    This file is part of Hastymail.

    Hastymail is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Hastymail is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hastymail; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* $Id: session_2.php,v 1.51 2005/12/07 16:46:57 sailfrog Exp $
*/

/**
 *  read in the contacts list from LDAP book defined in hastymail.conf
 *  @param      $config     array       configuration settings
 *  @param      $key        string      encrypted user password
 *  @param      $filter     string      search filter
 *  @return                 array       any LDAP contacts we could find
 *
 */
function hm_get_ldap_contacts($config, $key, $filter) {
    $contacts = array();
    if (!function_exists('ldap_connect')) {
        return $contacts;
    }
    if (isset($config['ldap_ssl']) && $config['ldap_ssl'] == 'true') {
        $ldap_server = 'ldaps://'.$config['ldap_server'];
    }
    else {
        $ldap_server = 'ldap://'.$config['ldap_server'];
    }
    $str = $_SESSION['str'];
    if ($ldapp= @ldap_connect($ldap_server, $config['ldap_port'])) {
        if (isset($config['ldap_auth']) &&
            $config['ldap_auth'] == 'true') {
            $ldap_rdn = $_SESSION['user'];
            $ldap_pass = hmw_crypt($_SESSION['pass'], $config, 'decrypt', $key);
            $bind = @ldap_bind($ldapp, $ldap_rdn, $ldap_pass);
        }
        else {
            $bind = @ldap_bind($ldapp);
        }
        if ($bind) {
            if ($filter) {
                $filter = str_replace(array('*', '(', ')', "\\", "\0"),
                          array('\0x2a', '\0x28', '\0x29', '\0x5c', '\0x00'), $filter);
                $search = '(|(|(cn=*'.$filter.'*) (mail=*'.$filter.
                          '*))(sn=*'.$filter.'*))';
            }
            elseif (isset($config['ldap_search_term']) &&
                $config['ldap_search_term'] != '') {
                $search = $config['ldap_search_term'];
            }
            else {
                $search = 'cn=*';
            }
            $ldap_attr = @ldap_search($ldapp, $config['ldap_base_dn'], $search);
            $info = ldap_get_entries($ldapp, $ldap_attr);
            ldap_close($ldapp);
            array_shift($info);
            $i = 0;
            foreach ($info as $array) {
                if (isset($array['mail'][0]) && strstr($array['mail'][0], '@')) {
                    $contacts[$i] = array( $i, trim($array['mail'][0]), 
                                    trim($array['cn'][0], ''));
                    $i++;
                }
            }
        }
        else {
            ldap_close($ldapp);
            $_SESSION['notices'][] = $str[197];
        }
    }
    else {
        $_SESSION['notices'][] = $str[198];
    }

    return $contacts;
}

/**
 *  read in the contacts list
 *
 *  @param      $user       string      username
 *  @param      $dir        string      settings dir
 *  @param      $filter     string      search filter
 *  @return                 array       array of personal contacts
 *
 */
function hm_get_contacts($user, $dir, $filter=false) {
    $contacts = array();
    if ($contacts_file = @fopen ($dir.strtolower($user).'.contacts', 'r')) {
        while (!feof ($contacts_file)) {
            $line = fgets($contacts_file, 1024);
            $parts = preg_split("/\|/", $line, 4);
            if (count($parts) == 4) {
                if ($filter) {
                   if (stristr($parts[2], $filter) || stristr($parts[1], $filter)) { 
                        $contacts[$parts[0]] = $parts;
                    }
                }
                else {
                    $contacts[$parts[0]] = $parts;
                }
            }
        }
    }
    return $contacts;
}

/**
 *  Read in the contacts list from DB
 *
 *  @param  string      $user       Username
 *  @param  array       $config     Configuration Array
 *  @param  string      $filter     Search Filter
 *
 */
function hm_get_db_contacts($user, $config, $filter=false) {
    $contacts = array();
    if (empty($config['contacts_dbdsn']) || empty($config['contacts_dbtable']) ||
        empty($config['contacts_dbuserfld']) || empty($config['contacts_dbdatafld']) ||
        empty($config['contacts_dbcontactfld'])) {
        $_SESSION['notices'][] = 'Database settings for contacts are incomplete. '.
                                 'Please contact your system administrator.';
        return $contacts;
    }
    include_once('DB.php');
    $db =& DB::Connect($config['contacts_dbdsn']);
    if (PEAR::isError($db)) {
        $_SESSION['notices'][] = 'Connection to "Contacts" database could not be '.
                                'established.  Please contact your system administrator.';
        return $contacts;
    }
    $sql = 'select * from %s where %s=%s';
    $sql = sprintf($sql , $config['contacts_dbtable'], $config['contacts_dbuserfld'],
                   $db->quoteSmart($user));
    $res =& $db->query($sql);
    if (PEAR::isError($res)) {
        $_SESSION['notices'][] = 'An error occured searching for contacts.';
        $_SESSION['notices'][] = $res->getMessage();
        return $contacts;
    }
    if ($res->numRows() == 0) {
        $db->disconnect();
        return $contacts;
    }
    while( $row =& $res->fetchRow(DB_FETCHMODE_ASSOC) ) {
        $contact_name = $row[$config['contacts_dbcontactfld']];
        $contact_data = unserialize($row[$config['contacts_dbdatafld']]);
        if (count($contact_data) == 4) {
            if ($filter) {
                if (stristr($contact_data[2], $filter) || stristr($contact_data[1],
                    $filter)) {
                    $contacts[$contact_data[0]] = $contact_data;
                }
            }
            else {
                $contacts[$contact_data[0]] = $contact_data;
            }
        }
    }
    return $contacts;
}

/**
 *  sort the contact list
 *
 *  @param      $contacts       array   the contacts array
 *  @return                     array   alpha sorted contacts
 *
 */
function hm_sort_contacts($contacts, $compose=false) {
    $i = 0;
    $new_contacts = array();
    if (!function_exists('hm_cmpsort')) {
        function hm_cmpsort ($a, $b) {
            $index = 2;
            if (isset($GLOBALS['_SESSION']['contact_sort'])) {
                $index = (int) intval($GLOBALS['_SESSION']['contact_sort']);
            }
            if ($index < 0 || $index > 3) {
                $index = 2;
            }
            return strcmp(strtolower($a[$index]), strtolower($b[$index]));
        }
    }
    usort($contacts, "hm_cmpsort");
    foreach ($contacts as $index => $array) {
        $new_contacts[$i] = $array;
        if ($compose) {
            $new_contacts[$i][0] = $i;
        }
        $i++;
    }
    return $new_contacts;
}

/**
 * Get user settings from DB
 *
 * @param   string  $user       username
 * @param   array   $config     configuration values
 */
function hm_get_db_settings($user, $config) {
    $settings = array();
    $profiles = array();
    if (!isset($_SESSION['str'])) {
        $str = array(
            366 => 'No database connection string specified for settings',
            367 => 'Unable to connect to database for settings.',
            368 => 'Unable to query database for settings. Database generated an error');
    }
    else {
        $str = $_SESSION['str'];
    }
    include_once('DB.php');
    if (empty($config['settings_dbdsn'])) {
        $_SESSION['notices'][] = $str[366];
        return $settings;
    }
    $db =& DB::Connect($config['settings_dbdsn'], array('debug'=>2));
    if (PEAR::isError($db)) {
        $_SESSION['notices'][] = $str[367];
        $_SESSION['notices'][] = $db->getMessage();
        return $settings;
    }
    $sql = sprintf('select * from %s where %s = %s', $config['settings_dbtable'], 
           $config['settings_table_user'], $db->quoteSmart($user));
    $res =& $db->query($sql);
    if (PEAR::isError($res)) {
        $_SESSION['notices'][] = $str[368];
        $_SESSION['notices'][] = $res->getMessage();
        $db->disconnect();
        return $settings;
    }
    $profile_count = 1;
    while( $row =& $res->fetchRow(DB_FETCHMODE_ASSOC) ) {
        switch( $row[$config['settings_table_key']] ) {
            case 'hidden':
            case 'namespace':
            case 'new_mail_folders':
                $tmparr = array();
                $tmparr = unserialize($row[$config['settings_table_data']]);
                if (!empty($tmparr) && is_array($tmparr)) {
                    foreach($tmparr as $data) {
                        $settings[] = $row[$config['settings_table_key']].'='.$data;
                    }
                }
                break;
            case 'folded':
                $folds = unserialize($row[$config['settings_table_data']]);
                if (!empty($folds) && is_array($folds)) {
                    foreach( $folds as $folder=>$value ) {
                        $settings[] = 'folded='.$folder;
                    }
                }
                break;
            case 'sort':
                $sort = unserialize($row[$config['settings_table_data']]);
                if (!empty($sort) && is_array($sort)) {
                    foreach($sort as $folder=>$order) {
                        $settings[] = 'sort='.$folder.'|'.$order;
                    }
                }
                break;
            case 'profiles':
                $profiles = unserialize($row[$config['settings_table_data']]);
                $_SESSION['profiles'] = $profiles;
                break;
            case 'profile_count':
                $profile_count = $row[$config['settings_table_data']];
                $settings[] = $row[$config['settings_table_key']].'='.
                             $row[$config['settings_table_data']];
                break;
            default:
                $settings[] = $row[$config['settings_table_key']].'='.
                              $row[$config['settings_table_data']];
        }
    }
    if (count($settings) == 0) {
        $_SESSION['first_login'] = true;
    }
    $db->disconnect();
    return $settings;
}

/**
 * Get the users settings from a file
 *
 * @param   string      $user       username
 * @param   array       $config     configuration settings
 */
function hm_get_file_settings($user, $config) {
    $settings = array();
    $set_file = $config['settings_dir'] . strtolower($user) . '.settings';
    if (!is_readable($set_file)) {
        $_SESSION['first_login'] = true;
    }
    if (is_readable($set_file)) {
        $settings = @file($set_file);
    }
    $prfcnt = false;
    foreach($settings as $id=>$data) {
        if (preg_match('/profile_count=(\d+)/', $data, $matches)) {
            $profile_count = $matches[1];
            $prfcnt = true;
            break;
        }
        else {
            $profile_count = 1;
        }
    }
    $data = array();
    if ($prfcnt == false) {
        $settings[] = 'profile_count=1';
    }
    $data = @file($config['settings_dir'].strtolower($user).'.profile');
    if (!isset($profile_count)) {
        $profile_count = 1;
    }
    for ($i=0;$i<$profile_count;$i++) {
        $_SESSION['profiles'][$i]['from_name'] = '';
        $_SESSION['profiles'][$i]['from_address'] = '';
        $_SESSION['profiles'][$i]['sig'] = '';
        /* bail if no data */
        if (!isset($data[0])) {
            continue;
        }
        else {
            /* get profile from name */
            $_SESSION['profiles'][$i]['from_name'] = trim($data[0]);
            array_shift($data);
            if (!isset($data[0])) {
                continue;
            }
            else {
                /* get profile from address */
                $_SESSION['profiles'][$i]['from_address'] = trim($data[0]);
                array_shift($data);
            }
        }
        /* get profile signature */
        if (isset($data[0]) && preg_match("/^\{(\d+)\}/", $data[0], $matches)) {
            if (isset($matches[1])) {
                $len = $matches[1];
                $data[0] = substr($data[0], (strlen($len) + 2));
                $sig = '';
                if ($len > 0) {
                    while ($len > 0) {
                        $sig .= $data[0];
                        $len -= strlen($data[0]);
                        array_shift($data);
                    }
                    $_SESSION['profiles'][$i]['sig'] = $sig;
                }
                else {
                    array_shift($data);
                }
            }
        }
        else {
            continue;
        }
    }
    return $settings;
}

/**
 *  get the users settings
 *
 *  @param    array     $data   settings passed from various systems
 *  @return             array       
 *
 */
function hm_get_settings($data) {
    $settings = array();
    if (!empty($data) && is_array($data)) {
        foreach($data as $id=>$line) {
            $pair = explode('=' , $line, 2);
            if (isset($pair[0]) && isset($pair[1])) {
                switch($pair[0]) {
                    case 'hidden':
                    case 'new_mail_folders':
                    case 'namespace':
                        $settings[$pair[0]][] = trim($pair[1]);
                        break;
                    case 'folded':
                        $settings['folded'][trim($pair[1])] = 1;
                        break;
                    case 'sort':
                        $parts = preg_split("/\|/", $pair[1], 2);
                        if (isset($parts[0]) && $parts[1]) {
                            $settings['sort'][$parts[0]] = trim($parts[1]);
                        }
                        break;
                    default:
                        $settings[$pair[0]] = trim($pair[1]);
                        break;
                }
            }
        }

    }
    return $settings;
}

/**
 *
 *  Write the contacts to a db
 *
 *  @param      string      $user       Username
 *  @param      array       $config     Configuration Array
 *  @param      array       $contacts   List of contacts
 *  @return     boolean                 True on success, false on failure
 *
 */
function hm_write_db_contacts( $user , $config, $contacts ) {
    $results = false;
    if (empty($config['contacts_dbdsn']) || empty($config['contacts_dbtable']) ||
        empty($config['contacts_dbuserfld']) || empty($config['contacts_dbdatafld'])
         || empty($config['contacts_dbcontactfld'])) {
        $_SESSION['notices'][] = 'Database settings for contacts are incomplete.  '.
                                 'Please contact your system administrator.';
        return $results;
    }
    include_once('DB.php');
    $db =& DB::Connect($config['contacts_dbdsn']);
    if (PEAR::isError($db)) {
        $_SESSION['notices'][] = 'Connection to "Contacts" database could not be '.
                                 'established.  Please contact your system '.
                                 'administrator.';
        return $results;
    }
    $d_sql = 'delete from %s where %s=%s';
    $w_sql = 'insert into %s (%s, %s, %s) values (%s, %s, %s)';
    $sql = sprintf($d_sql , $config['contacts_dbtable'], $config['contacts_dbuserfld'],
                   $db->quoteSmart($user));
    $res =& $db->query($sql);
    if (PEAR::isError($res)) {
        $_SESSION['notices'][] = 'Unable to update "Contacts" data.  Please contact your'.
                                 ' system administrator(1).';
        $db->disconnect();
        return $results;
    }
    $i = 0;
    foreach( $contacts as $array ) {
        $contact_name = $array[1];
        $contact_data = $array;
        $contact_data[0] = $i;
        $contact_data = serialize($contact_data);
        $sql = '';
        $sql = sprintf($w_sql , $config['contacts_dbtable'],
                       $config['contacts_dbuserfld'], $config['contacts_dbcontactfld'],
                       $config['contacts_dbdatafld'], $db->quoteSmart($user),
                       $db->quoteSmart($contact_name), $db->quoteSmart($contact_data));
        $res =& $db->query($sql);
        if (PEAR::isError($res)) {
            $_SESSION['notices'][] = 'Unable to update "Contacts" data.  Please contact '.
                                     'your system administrator(2).';
            $_SESSION['notices'][] = $res->getMessage();
            $db->disconnect();
            return $results;
        }
        $i++;
    }
    $db->disconnect();
    $results = true;
    return $results;
}

/**
 *  write the contacts to file
 *
 *  @param      $user       string      username
 *  @param      $dir        string      settings directory
 *  @param      $contacts   array       the contacts array
 *  @return                 boolean     true on success false on failure
 *
 */
function hm_write_contacts($user, $dir, $contacts) {
    $result = false;
    $contacts = hm_sort_contacts($contacts);
    if ($contacts_file = @fopen ($dir.strtolower($user).'.contacts', 'w')) {
        $i = 0;
        foreach ($contacts as $array) {
            $line = $i.'|'.trim($array[1]).'|'.trim($array[2]).'|'.trim($array[3]);
            str_replace("\n", '', $line);
            fputs($contacts_file, trim($line)."\n");
            $i++;
        }
        fclose($contacts_file);
        $result = true;
    }
    return $result;
}

/**
 *  write the users settings to a db
 *
 *  @param  string      $user       username
 *  @param  array       $config     configuration array
 *  @return     void
 */
function hm_write_db_settings($user, $config) {
    include_once('DB.php');
    $settings = $_SESSION['settings'];
    if (isset($_SESSION['profiles'])) {
        $profiles = $_SESSION['profiles'];
    }
    else {
        $profiles = array();
    }
    $str = $_SESSION['str'];
    if (empty($config['settings_dbdsn'])) {
        $_SESSION['notices'][] = $str[366];
        return;
    }
    $db =& DB::Connect($config['settings_dbdsn'], array('debug'=>2));
    if (PEAR::isError($db)) {
        $_SESSION['notices'][] = $str[367];
        $_SESSION['notices'][] = $db->getMessage();
        return;
    }
    $dbset = array();
    foreach($settings as $key=>$value) {
        if (is_array($value)) {
            $value = serialize($value);
            $dbset[$key] = $value;
        } else {
            $dbset[$key] = $value;
        }
    }
    if (!empty($profiles)) {
        $dbset['profiles'] = serialize($profiles);
    }
    foreach($dbset as $key=>$data) {
        $clean_sql = sprintf('delete from %s where %s=%s and %s=%s',
                        $config['settings_dbtable'],
                        $config['settings_table_user'],
                        $db->quoteSmart($user),
                        $config['settings_table_key'],
                        $db->quoteSmart($key));
        $res =& $db->query($clean_sql);
        if (PEAR::isError($res)) {
            $_SESSION['notices'][] = $str[369];
            $_SESSION['notices'][] = $res->getMessage();
            break;
        }
        $set_sql = sprintf('insert into %s (%s,%s,%s) values (%s,%s,%s)',
                        $config['settings_dbtable'],
                        $config['settings_table_user'],
                        $config['settings_table_key'],
                        $config['settings_table_data'],
                        $db->quoteSmart($user),
                        $db->quoteSmart($key),
                        $db->quoteSmart($data));
        $res =& $db->query($set_sql);
        if (PEAR::isError($res)) {
            $_SESSION['notices'][] = $str[369];
            $_SESSION['notices'][] = $res->getMessage();
            break;
        }
    }
    $db->disconnect();
}

/**
 *  write the users settings to file
 *
 *  @param      $user       string      username
 *  @param      $dir        string      settings dir
 *  @return                 void
 * 
 */  
function hm_write_file_settings($user, $dir) {
    $settings = $_SESSION['settings'];
    $str = $_SESSION['str'];
    $profile_count = $settings['profile_count'];
    if (isset($settings['sort'])) {
        $sorts = $settings['sort'];
        $settings['sort'] = array();
        foreach($sorts as $index => $value) {
            $settings['sort'][] = $index.'|'.$value;
        }
    }
    if (isset($settings['folded'])) {
        $folds = $settings['folded'];
        $settings['folded'] = array();
        foreach ($folds as $index => $val) {
            $settings['folded'][] = $index;
        }
    }
    $settings_file = @fopen ($dir.strtolower($user).'.settings', 'w');
    if (is_resource($settings_file)) {
        foreach ($settings as $index => $value) {
            if (is_array($value)) {
                foreach ($value as $element) {
                    fputs($settings_file, $index."=".$element."\n");
                }
            }
            else { 
                fputs($settings_file, $index."=".$value."\n");
            }
        }
        fclose($settings_file);
    }
    else {
        $_SESSION['notices'][] = $str[272];
    }
    $profile_file = @fopen ($dir.strtolower($user).'.profile', 'w');
    if (!isset($_SESSION['profiles'])) {
        $_SESSION['profiles'][0]['from_name'] = '';
        $_SESSION['profiles'][0]['from_address'] = '';
        $_SESSION['profiles'][0]['sig'] = '';
    }
    if (is_resource($profile_file)) {
        for ($i=0;$i<$profile_count;$i++) {
            if ($profile_file) {
               fputs($profile_file, trim($_SESSION['profiles'][$i]['from_name'])."\n");
               fputs($profile_file, trim($_SESSION['profiles'][$i]['from_address'])."\n");
               fputs($profile_file, '{'.
                    strlen(rtrim($_SESSION['profiles'][$i]['sig'])).'}'.
                    rtrim($_SESSION['profiles'][$i]['sig'])."\n");
            }
        }
    fclose($profile_file);
    }
    else {
        $_SESSION['notices'][] = $str[273];
    }
}

/**
 * check for missing special folders
 *
 *  @return             void
 */
function hm_check_folders($imap_stream, $quiet=false) {
    $write = false;
    $str = $_SESSION['str'];
    $specials = array(299 => 'default_folder', 27  => 'trash_folder', 
                      28  => 'sent_folder',    209 => 'drafts_folder');
    foreach ($specials as $index => $val) {
        if (isset($_SESSION['settings'][$val]) &&
            !in_array(urldecode($_SESSION['settings'][$val]), $_SESSION['folders'])) {
            unset($_SESSION['settings'][$val]);
            $write = true;
            if (!$quiet) {
                $_SESSION['notices'][] = $str[$index];
            }
        }
    }
    if ($write) {
        hmw_write_settings($_SESSION['user'], $_SESSION['config']);
    }
    if (isset($_SESSION['missing_folders_flag']) && 
        $_SESSION['missing_folders_flag'] == 1 && 
        is_array($_SESSION['missing_folders'])) {
        $tmp = $str[47].'<br>';
        foreach ($_SESSION['missing_folders'] as $missing) {
            $tmp .= "<b>$missing</b><br>";
        }
        $tmp .= $str[49];
        $_SESSION['notices'][] = $tmp;
        $_SESSION['missing_folders_flag'] = 0;
        unset($_SESSION['missing_folders_flag']);
        $_SESSION['folders'] = hm_boxlist($imap_stream, true, 'all');
    }
    return;
}    
?>
